# -*- coding: utf-8 -*-
# @Time    : 04/09/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : data_preprocessing.py
# @Software: PyCharm
import os
import glob
import pandas as pd
import numpy as np
from openpy_fxts.preprocessing_data.utils import add_time_features
from openpy_fxts.models.clustering.prepare_data import get_scenarios_for_time_features
from openpy_fxts.error_handling_logging import update_logg_file

import seaborn as sns
import matplotlib.pyplot as plt


def _user_id(data):
    return data['Unnamed: 1'][1]


def _clean_file(df):
    df = df.drop([0, 1, 2, 3, 4, 5, 6])
    df = df.dropna(how='all', axis=1)
    df.columns = list(df.iloc[:1].values[0])
    df = df.drop([df.index[0]], axis=0)
    df.set_index(df.columns[0], inplace=True)
    df = df.loc[pd.notnull(df.index)]
    df = df[[x for x in list(df.columns) if str(x) != 'nan']]
    df['W'] = df['Wh'] / 0.25
    return df[list(df.columns)[-2:]]


def dimensionality_reduction(df):
    df_clean = pd.DataFrame()  # datasets frame vacío.
    for h in range(0, 24):  # h va a ieterar 24 veces representando las 24 horas del día.
        df_clean = pd.DataFrame()
        k = [h]
        Hora_Dia = df[df['hour'].isin(k)]  # filtramos por hora del dia
        Hora_Dia_Acumu = Hora_Dia.groupby(['year', 'mount', 'Dia_M', 'hour'])[
            'E_Activa(Wh)'].sum()  # Sumar las mediciones de los periodos de 15 min entre horas
        Hora_Dia_Acum_ = pd.DataFrame(pd.DataFrame(Hora_Dia_Acumu))  # convertir a datasets frame a dataframe
        Hora_Dia_Acum_ = Hora_Dia_Acum_.reset_index(drop=True)  # se elimina niveles de índices
        vector = Hora_Dia_Acum_['E_Activa(Wh)']
        name_colum = '$%d$' % h  # nombre de la columna
        df_clean = df_clean.copy()
        df_clean[name_colum] = vector  # coloca el vector hora h en la columna h de df_celan
        df_clean = pd.concat([df_clean, df_clean], axis=1)  # adjuntos el nuevo vector de las horas Al dataframe
        df_clean = df_clean.dropna(how='any')  # Elimino datos null de df
        df_clean.columns = range(
            df_clean.shape[1])  # cambiamos el tipo de datos de los nombres de las columnas para redireccionar luego

    return df_clean


class read_BBDD:

    def multiple_user_files(
            self,
            file_folder_path: str = None,
            save_folder_path: str = None
    ):
        os.chdir(file_folder_path)
        filelist = glob.glob('*.xlsx')  # Muestra una lista de todos los archivos excel dentro del directorio
        i = 0
        df_user_Max_W = pd.DataFrame()
        df_user_W = pd.DataFrame()
        for filename in filelist:
            print(filename)
            data_user = pd.read_excel(filename)
            user = _user_id(data_user)
            data_user = _clean_file(data_user)
            df_aux1, df_aux2 = pd.DataFrame(), pd.DataFrame()
            df_aux1[user] = data_user[data_user.columns[0]]
            df_aux2[user] = data_user[data_user.columns[1]]
            df_user_Max_W = pd.concat([df_user_Max_W, df_aux1], axis=1)
            df_user_W = pd.concat([df_user_W, df_aux2], axis=1)
        df_user_Max_W.to_csv(f'{save_folder_path}\\alluser_Max_W.csv')
        df_user_W.to_csv(f'{save_folder_path}\\alluser_W.csv')


class scenarios:

    def multiple_users(
            self,
            file_path,
            index_col: int = 0,
            parse_dates: bool = True,
            dict_scenario: dict = None,
            id_col_users:  str = 'nroserie',
            variable: str = 'kW',
            col_user: str = None,
            col_meas: list = None
    ):
        data = pd.read_csv(
            file_path,
            index_col=index_col,
            parse_dates=parse_dates
        )

        data = pd.pivot_table(
            data=data,
            index=data.index,
            columns=id_col_users,
            values=variable
        )
        data.index.name = 'index'
        if data.isnull().sum().sum() > 0:
            moving_window_replace(data.values, n_days=7)
        data = add_time_features(data)
        if dict_scenario is None:
            scenarios = get_scenarios_for_time_features(
                data,
                seasons=True,
                month=True,
                day_name=True,
                day_type=True
            )
            return scenarios
        else:
            for key, value in dict_scenario.items():
                if value is None:
                    dict_scenario[key] = list(data[key].unique())
            data_filter = data[
                (data.seasons.isin(dict_scenario['seasons'])) &
                (data.month.isin(dict_scenario['month'])) &
                (data.year.isin(dict_scenario['year'])) &
                (data.day_name.isin(dict_scenario['day_name'])) &
                (data.day_type.isin(dict_scenario['day_type']))
                ]
            if data_filter.empty:
                print('Modify the scenario filtering dictionary')
                return None
            else:
                data_filter
                return data_filter

    def single_user(
            self,
            file_path,
            dict_scenario: dict = None,
            id_user: str = None,
            only_df: bool = None,
            index_col: int = 0,
            parse_dates: bool = True

    ):
        if dict_scenario is not None:
            only_df = False
        data = pd.read_csv(
            file_path,
            index_col=index_col,
            parse_dates=parse_dates
        )
        if data.isnull().sum().sum() > 0:
            moving_window_replace(data.values, n_days=7)
        if id_user is None:
            col_init = list(data.columns)[0]
            data = data[[col_init]]
        else:
            #col_init = id_user
            data = data[[id_user]]
        data = add_time_features(data)
        if dict_scenario is None:
            if only_df:
                return data
            else:
                return get_scenarios_for_time_features(
                    data,
                    seasons=True,
                    month=True,
                    day_name=True,
                    day_type=True
                )
        else:
            for key, value in dict_scenario.items():
                if value is None:
                    dict_scenario[key] = list(data[key].unique())
            data_filter = data[
                (data.seasons.isin(dict_scenario['seasons'])) &
                (data.month.isin(dict_scenario['month'])) &
                (data.year.isin(dict_scenario['year'])) &
                (data.day_name.isin(dict_scenario['day_name'])) &
                (data.day_type.isin(dict_scenario['day_type']))
            ]
            if data_filter.empty:
                print('Modify the scenario filtering dictionary')
                return None
            else:
                data_filter
                return data_filter

    def filter_scenario(self):
        pass


def moving_window_replace(
        data: np.array = None,
        res_min: int = 15,
        n_days: int = None,
        previous: bool = None,
        back: bool = None
):
    if previous is None and back is None:
        previous = True
    window = (24 * 60) / res_min
    if n_days is not None:
        window = window * n_days
    n = 0
    window = int(window)
    while len(data[np.isnan(data)]) > 0:
        print(f'iter:{n}, n_NaN:{len(data[np.isnan(data)])}')
        for row in range(data.shape[0]):
            for col in range(data.shape[1]):
                if np.isnan(data[row, col]):
                    if previous:
                        try:
                            data[row, col] = data[row - window, col]
                        except IndexError:
                            try:
                                data[row, col] = data[row + window, col]
                            except IndexError:
                                pass
                    if back:
                        try:
                            data[row, col] = data[row + window, col]
                        except IndexError:
                            try:
                                data[row, col] = data[row - window, col]
                            except IndexError:
                                pass
        n += 1


def moving_window_mean(
        data: np.array = None,
        n_hours: int = None,
        res_min: int = None,
        n_past: bool = None,
        n_future: bool = None,
):
    aux = 0
    if n_hours is None:
        n = 1
        n_hours = n * 60
    while len(data[np.isnan(data)]) > 0:
        print(f'iter:{aux}, n_NaN:{len(data[np.isnan(data)])}')
        if n_past is None and n_future is None and res_min is not None:
            n_past, n_future = n_hours / res_min, n_hours / res_min
        elif n_past is None and n_future is None and res_min is None:
            n_past, n_future = 4, 4
        for row in range(data.shape[0]):
            for col in range(data.shape[1]):
                if np.isnan(data[row, col]):
                    data[row, col] = np.concatenate(
                        (
                            data[row - n_past: row, col],
                            data[row: row + n_future, col]
                        )
                    ).mean()
                    if np.isnan(data[row, col]):
                        data[row, col] = data[row - n_past: row, col].mean()
                        if np.isnan(data[row, col]):
                            data[row, col] = data[row: row + n_future, col].mean()
        aux += 1


class segmentation_analysis:

    def add_discrete_variables(self, season, mount, type, ):
        pass


class dimensionality_reduction:
    pass



import tensorflow as tf

from openpy_fxts.architectures.get_arch_mdls import get_architecture
from openpy_fxts.preprocessing_data.forecast.prepare_data import pre_processing_data
from openpy_fxts.architectures.utils import mdl_characteristics
from openpy_fxts.architectures.utils import callbacks
from openpy_fxts.architectures.utils import learning_curve
from openpy_fxts.architectures.utils import process_values_preliminary
from time import time

tkm = tf.keras.models
tkl = tf.keras.layers
tkloss = tf.keras.losses
tko = tf.keras.optimizers
tku = tf.keras.utils

sim_aux = {
    'view_summary': False,  # Summary for architecture select
    'metrics': True,
    'preliminary': False,
    'plt_model': False,  # Don't change to True, need review
    'plt_history': True,
    'plt_results': True,
    'plt_metrics': False,  # Don't change to True, need review
    'view_plots': False,
    'verbose': 1,
    'patience': 5,
    'time_init': time()
}

mdl_aux = {
    'optimizer': 'adam',
    'loss': 'mse',
    'metrics': ['accuracy'],  # Metrics to train
    'batch_size': 128,  # Batch size for training. { [32, 64],[*128, 256] }- Good starters
    'epochs': 100,  # Number of epochs to train for.
}


class base_model:

    def __init__(
            self,
            config_data=None,
            config_mdl=None,
            config_sim=None,
            config_arch=None,
            name_mdl=None,
            type_mdl=None,
            app=None
    ):
        if config_sim is None:
            config_sim = sim_aux
        if config_mdl is None:
            config_mdl = mdl_aux
        # Parameters for dataset
        self.config_data = config_data
        self.n_past = config_data['n_past']
        self.n_future = config_data['n_future']
        self.n_inp_ft = config_data['n_inp_ft']
        self.n_out_ft = config_data['n_out_ft']
        # Parameters for model
        self.config_mdl = config_mdl
        self.optimizer = config_mdl['optimizer']
        self.loss = config_mdl['loss']
        self.metrics = config_mdl['metrics']
        self.batch_size = config_mdl['batch_size']  # Batch size for training.
        self.epochs = config_mdl['epochs']  # Number of epochs to train for.
        # Parameters for simulation
        self.config_sim = config_sim
        self.verbose = config_sim['verbose']
        self.patience = config_sim['patience']
        self.plot_history = config_sim['plt_history']
        self.preliminary = config_sim['preliminary']
        self.config_arch = config_arch
        self.name_mdl = name_mdl
        self.type_mdl = type_mdl
        self.app = app

    def _architecture_model(self):
        get_arch = get_architecture(
            name_mdl=self.name_mdl,
            type_mdl=self.type_mdl,
            n_past=self.n_past,
            n_future=self.n_future,
            n_inp_ft=self.n_inp_ft,
            n_out_ft=self.n_out_ft,
            config_arch=self.config_arch,
            app=self.app
        )
        model = get_arch.select_model()
        return model

    def reshape_train_valid_for_fxts(self, pre_processed):
        # Training
        X_train = pre_processed['train']['X']
        y_train = pre_processed['train']['y']
        # Validation
        X_valid = pre_processed['valid']['X']
        y_valid = pre_processed['valid']['y']
        if self.type_mdl == 'MLP':
            if self.name_mdl == 'MLP_Dense':
                X_train_list, X_valid_list = [], []
                for i in range(X_train.shape[2]):
                    X_train_list.append(X_train[:, :, i])
                    X_valid_list.append(X_valid[:, :, i])
                return X_train_list, y_train, X_valid_list, y_valid
            if self.name_mdl == 'MLP_RNN':
                X_train_list, X_valid_list = [], []
                for i in range(X_train.shape[2]):
                    X_train_list.append(X_train[:, :, i])
                    X_valid_list.append(X_valid[:, :, i])
                return X_train_list, X_valid_list, X_valid, y_valid
        if self.type_mdl == 'Conv1D':
            if self.name_mdl == 'TimeDist_Conv1D_RNN':
                X_train = X_train.reshape(X_train.shape[0], X_train.shape[1], X_train.shape[2], 1)
                X_valid = X_valid.reshape(X_valid.shape[0], X_valid.shape[1], X_valid.shape[2], 1)
                return X_train, y_train, X_valid, y_valid
            else:
                return X_train, y_train, X_valid, y_valid
        else:
            return X_train, y_train, X_valid, y_valid

    def _training_mdl(self, filepath):
        pre_processed = pre_processing_data(
            self.config_data,
            train=True,
            valid=True
        ).transformer_data(
            dropnan=True
        )
        model = self._architecture_model()
        print(self.name_mdl)
        model.compile(
            optimizer=self.optimizer,
            loss=self.loss,
            metrics=self.metrics
        )
        mdl_characteristics(
            model,
            self.config_sim,
            filepath
        )
        X_train, y_train, X_valid, y_valid = self.reshape_train_valid_for_fxts(pre_processed)
        history = model.fit(
            x=X_train,
            y=y_train,
            validation_data=(
                X_valid,
                y_valid
            ),
            batch_size=self.batch_size,
            epochs=self.epochs,
            verbose=self.verbose,
            callbacks=callbacks(
                filepath,
                weights=True,
                verbose=self.verbose,
                patience=self.patience
            )
        )
        if self.plot_history:
            learning_curve(
                history=history,
                name_mdl=self.name_mdl,
                time_start=self.config_sim['time_init'],
                view_plots=self.config_sim['view_plots'],
                filepath=filepath
            )
        if self.preliminary:
            return process_values_preliminary(
                model=model,
                dict_test=pre_processing_data(
                    self.config_data,
                    test=True
                ).transformer_data(
                    dropnan=True
                ),
                config_data=self.config_data,
                config_sim=self.config_sim,
                name_mdl=self.name_mdl,
                type_mdl=self.type_mdl,
                app=self.app
            ).get_values(filepath)
        else:
            return None

    def _prediction_model(self, model_train, filepath):
        return process_values_preliminary(
            model=model_train,
            dict_test=pre_processing_data(
                self.config_data,
                test=True
            ).transformer_data(
                dropnan=True
            ),
            config_data=self.config_data,
            config_sim=self.config_sim,
            name_mdl=self.name_mdl,
            type_mdl=self.type_mdl,
            app=self.app
        ).get_values(filepath)


class base_class(base_model):

    def __init__(
            self,
            config_data=None,
            config_mdl=None,
            config_sim=None,
            config_arch=None,
            name_mdl=None,
            type_mdl=None,
            app=None
    ):
        super().__init__(config_data, config_mdl, config_sim, config_arch, name_mdl, type_mdl, app)

    def _build_mdl(self):
        return self._architecture_model()

    def training_mdl(
            self,
            filepath: str = None
    ):
        return self._training_mdl(filepath)

    def prediction_mdl(self, filepath: str = None):
        return self._prediction_model(self._build_mdl(), filepath)



# -*- coding: utf-8 -*-
# @Time    : 14/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm
from openpy_fxts.baseline_mdls import base_class

import tensorflow as tf

from openpy_fxts.base_tf import tkm, tkl
from openpy_fxts.preprocessing_data.forecast.prepare_data import pre_processing_data
import keras.utils.vis_utils
from importlib import reload

reload(keras.utils.vis_utils)
from openpy_fxts.architectures.utils import callbacks, learning_curve
from openpy_fxts.architectures.utils import _values_preliminary_2D
from openpy_fxts.architectures.utils import mdl_characteristics

tkl = tf.keras.layers
tko = tf.keras.optimizers
tkm = tf.keras.models
tkloss = tf.keras.losses
tku = tf.keras.utils
tkr = tf.keras.regularizers


class Seq2Seq_RNN_class(base_class):

    def __init__(
            self,
            config_data=None,
            config_mdl=None,
            config_sim=None,
            config_arch=None,
            name_mdl='Seq2Seq_LSTM',
            type_mdl='seq2seq',
    ):
        super().__init__(config_data, config_mdl, config_sim, config_arch, name_mdl, type_mdl)
        return


class Seq2Seq_RNN2_class(base_class):

    def __init__(
            self,
            config_data=None,
            config_mdl=None,
            config_sim=None,
            config_arch=None,
            name_mdl='Seq2Seq_LSTM2',
            type_mdl='seq2seq',
    ):
        super().__init__(config_data, config_mdl, config_sim, config_arch, name_mdl, type_mdl)
        return


class Seq2Seq_RNN_Batch_Drop_class(base_class):

    def __init__(
            self,
            config_data=None,
            config_mdl=None,
            config_sim=None,
            config_arch=None,
            name_mdl='Seq2Seq_LSTM_Batch_Drop',
            type_mdl='seq2seq',
    ):
        super().__init__(config_data, config_mdl, config_sim, config_arch, name_mdl, type_mdl)
        return


class Seq2Seq_BiRNN_class(base_class):

    def __init__(
            self,
            config_data=None,
            config_mdl=None,
            config_sim=None,
            config_arch=None,
            name_mdl='Seq2Seq_BiLSTM',
            type_mdl='seq2seq',
    ):
        super().__init__(config_data, config_mdl, config_sim, config_arch, name_mdl, type_mdl)
        return


class Seq2Seq_BiRNN2_class(base_class):

    def __init__(
            self,
            config_data=None,
            config_mdl=None,
            config_sim=None,
            config_arch=None,
            name_mdl='Seq2Seq_BiLSTM2',
            type_mdl='seq2seq',
    ):
        super().__init__(config_data, config_mdl, config_sim, config_arch, name_mdl, type_mdl)
        return


class Seq2Seq_Conv1D_BiRNN_class(base_class):

    def __init__(
            self,
            config_data=None,
            config_mdl=None,
            config_sim=None,
            config_arch=None,
            name_mdl='Seq2Seq_Conv1D_BiLSTM',
            type_mdl='seq2seq',
    ):
        super().__init__(config_data, config_mdl, config_sim, config_arch, name_mdl, type_mdl)
        return


class Seq2Seq_Multi_Head_Conv1D_BiRNN_class(base_class):

    def __init__(
            self,
            config_data=None,
            config_mdl=None,
            config_sim=None,
            config_arch=None,
            name_mdl='Seq2Seq_Multi_Head_Conv1D_BiLSTM',
            type_mdl='seq2seq',
    ):
        super().__init__(config_data, config_mdl, config_sim, config_arch, name_mdl, type_mdl)
        return


class Seq2Seq_BiRNN_with_Attention_class(base_class):

    def __init__(
            self,
            config_data=None,
            config_mdl=None,
            config_sim=None,
            config_arch=None,
            name_mdl='Seq2Seq_BiLSTM_with_Attention',
            type_mdl='seq2seq',
    ):
        super().__init__(config_data, config_mdl, config_sim, config_arch, name_mdl, type_mdl)
        return


class Seq2Seq_RNN_with_Luong_Attention_class(base_class):

    def __init__(
            self,
            config_data=None,
            config_mdl=None,
            config_sim=None,
            config_arch=None,
            name_mdl='Seq2Seq_LSTM_with_Luong_Attention',
            type_mdl='seq2seq',
    ):
        super().__init__(config_data, config_mdl, config_sim, config_arch, name_mdl, type_mdl)
        return


class Seq2Seq_Conv2D_RNN_class:

    def __init__(self, config_data=None, config_mdl=None, config_sim=None):
        self.config_data = config_data
        self.n_past = config_data['n_past']
        self.n_future = config_data['n_future']
        self.n_inp_ft = config_data['n_inp_ft']
        self.n_out_ft = config_data['n_out_ft']
        # Parameters for model
        self.config_mdl = config_mdl
        self.optimizer = config_mdl['optimizer']
        self.loss = config_mdl['loss']
        self.metrics = config_mdl['metrics']
        self.batch_size = config_mdl['batch_size']  # Batch size for training.
        self.epochs = config_mdl['epochs']  # Number of epochs to train for.
        self.units = config_mdl['units']  # no of lstm units
        self.dropout = config_mdl['dropout']
        # Parameters for simulation
        self.config_sim = config_sim
        self.verbose = config_sim['verbose']
        self.patience = config_sim['patience']
        self.plot_history = config_sim['plt_history']
        self.preliminary = config_sim['preliminary']
        self.name_model = 'Seq2Seq_ConvLSTM2D'

    def build_model(self):
        model = tkm.Sequential()
        model.add(tkl.BatchNormalization(name='batch_norm_0', input_shape=(self.n_past, self.n_inp_ft, 1, 1)))
        model.add(tkl.ConvLSTM2D(
            name='conv_lstm_1',
            filters=64,
            kernel_size=(10, 1),
            padding='same',
            return_sequences=True)
        )
        model.add(tkl.Dropout(0.2, name='dropout_1'))
        model.add(tkl.BatchNormalization(name='batch_norm_1'))
        model.add(tkl.ConvLSTM2D(
            name='conv_lstm_2',
            filters=64,
            kernel_size=(5, 1),
            padding='same',
            return_sequences=False)
        )
        model.add(tkl.Dropout(0.1, name='dropout_2'))
        model.add(tkl.BatchNormalization(name='batch_norm_2'))
        model.add(tkl.Flatten())
        # Repeat Vector
        model.add(tkl.RepeatVector(self.n_future))
        #
        if (self.n_inp_ft - self.n_out_ft) == 0:
            aux = 1
        else:
            aux = self.n_inp_ft - self.n_out_ft
        model.add(tkl.Reshape((self.n_future, self.n_out_ft, aux, 64)))
        model.add(tkl.ConvLSTM2D(
            name='conv_lstm_3',
            filters=64,
            kernel_size=(10, 1),
            padding='same',
            return_sequences=True)
        )
        model.add(tkl.Dropout(0.1, name='dropout_3'))
        model.add(tkl.BatchNormalization(name='batch_norm_3'))
        model.add(tkl.ConvLSTM2D(
            name='conv_lstm_4',
            filters=64,
            kernel_size=(5, 1),
            padding='same',
            return_sequences=True)
        )
        model.add(tkl.TimeDistributed(tkl.Dense(units=1, name='dense_1', activation='relu')))
        # model.add(Dense(units=1, name = 'dense_2'))
        return model

    def train_model(
            self,
            filepath: str = None
    ):
        data = pre_processing_data(self.config_data, train=True, valid=True)
        pre_processed = data.transformer_data()
        model = Seq2Seq_Conv2D_RNN_class(self.config_data, self.config_mdl).build_model()
        model.compile(
            optimizer=self.optimizer,
            loss=self.loss,
            metrics=self.metrics
        )
        mdl_characteristics(model, self.config_sim, filepath)
        # Training
        X_train = pre_processed['train']['X']
        y_train = pre_processed['train']['y']
        X_train = X_train.reshape_out(X_train.shape[0], X_train.shape[1], X_train.shape[2], 1, 1)
        y_train = y_train.reshape_out(y_train.shape[0], y_train.shape[1], y_train.shape[2], 1, 1)

        X_valid = pre_processed['valid']['X']
        y_valid = pre_processed['valid']['y']
        X_valid = X_valid.reshape_out(X_valid.shape[0], X_valid.shape[1], X_valid.shape[2], 1, 1)
        y_valid = y_valid.reshape_out(y_valid.shape[0], y_valid.shape[1], y_valid.shape[2], 1, 1)

        history = model.fit(
            X_train,
            y_train,
            epochs=self.epochs,
            validation_data=(
                X_valid,
                y_valid
            ),
            batch_size=self.batch_size,
            verbose=self.verbose,
            callbacks=callbacks(filepath, weights=True)
        )
        if self.config_data['plt_history']:
            learning_curve(history, self.name_model, self.config_data['time_init'], filepath=filepath)
        if self.config_data['preliminary']:
            data = pre_processing_data(self.config_data, test=True)
            dict_test = data.transformer_data()
            _values_preliminary_2D(model, dict_test, self.config_data)
        return model

    def prediction(
            self,
            model
    ):
        data = pre_processing_data(self.config_data, test=True)
        dict_test = data.transformer_data()
        yhat = _values_preliminary_2D(model, dict_test, self.config_data)
        return yhat



# -*- coding: utf-8 -*-
# @Time    : 29/08/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

import pandas as pd
import numpy as np


def get_scenarios_for_time_features(
        data,
        seasons: bool = False,
        month: bool = False,
        day_name: bool = False,
        day_type: bool = False,
):
    scenarios = dict()
    aux = 0
    if not seasons and not month and not day_name and not day_type:
        return None

    if not seasons and month and not day_name and not day_type:
        for month in list(data['month'].unique()):
            data_filter = data[data['month'] == month].copy()
            if not data_filter.empty:
                data_filter['sce'] = aux
                scenarios[month] = data_filter
                aux += 1
        return scenarios

    if not seasons and month and day_name and not day_type:
        for month in list(data['month'].unique()):
            for day_name in list(data['day_name'].unique()):
                data_filter = data[(data['month'] == month) & (data['day_name'] == day_name)].copy()
                if not data_filter.empty:
                    data_filter['sce'] = aux
                    scenarios[f'{month}_{day_name}'] = data_filter
                    aux += 1
        return scenarios

    if not seasons and month and day_name and day_type:
        for month in list(data['month'].unique()):
            for day_name in list(data['day_name'].unique()):
                df_aux = pd.DataFrame()
                for day_type in list(data['day_type'].unique()):
                    data_filter = data[(data['month'] == month) & (data['day_name'] == day_name) & (data['day_type'] == day_type)].copy()
                    if not data_filter.empty:
                        data_filter['sce'] = aux
                        df_aux = pd.concat([df_aux, data_filter])
                        aux += 1
                        scenarios[f'{month}_{day_name}_{day_type}'] = df_aux
        return scenarios

    if not seasons and not month and day_name and not day_type:
        for day_name in list(data['day_name'].unique()):
            data_filter = data[data['day_name'] == day_name].copy()
            if not data_filter.empty:
                data_filter['sce'] = aux
                scenarios[day_name] = data_filter
                aux += 1
        return scenarios

    if not seasons and not month and day_name and day_type:
        for day_name in list(data['day_name'].unique()):
            df_aux = pd.DataFrame()
            for day_type in list(data['day_type'].unique()):
                data_filter = data[(data['day_name'] == day_name) & (data['day_type'] == day_type)].copy()
                if not data_filter.empty:
                    data_filter['sce'] = aux
                    df_aux = pd.concat([df_aux, data_filter])
                    aux += 1
                    scenarios[f'{day_name}_{day_type}'] = df_aux
        return scenarios

    if not seasons and not month and not day_name and day_type:
        for day_type in list(data['day_type'].unique()):
            data_filter = data[data['day_type'] == day_type].copy()
            if not data_filter.empty:
                data_filter['sce'] = aux
                scenarios[day_type] = data_filter
                aux += 1
        return scenarios

    if seasons and not month and not day_name and not day_type:
        for season in list(data['seasons'].unique()):
            data_filter = data[data['seasons'] == season].copy()
            if not data_filter.empty:
                data_filter['sce'] = aux
                scenarios[season] = data_filter
                aux += 1
        return scenarios

    if seasons and month and not day_name and not day_type:
        for season in list(data['seasons'].unique()):
            for month in list(data['month'].unique()):
                data_filter = data[(data['seasons'] == season) & (data['month'] == month)].copy()
                if not data_filter.empty:
                    data_filter['sce'] = aux
                    scenarios[f'{season}_{month}'] = data_filter
                    aux += 1
        return scenarios

    if seasons and month and day_name and not day_type:
        for season in list(data['seasons'].unique()):
            for month in list(data['month'].unique()):
                for day_name in list(data['day_name'].unique()):
                    data_filter = data[(data['seasons'] == season) & (data['month'] == month) & (data['day_name'] == day_name)].copy()
                    if not data_filter.empty:
                        data_filter['sce'] = aux
                        scenarios[f'{season}_{month}_{day_name}'] = data_filter
                        aux += 1
        return scenarios

    if seasons and month and day_name and day_type:
        for season in list(data['seasons'].unique()):
            for month in list(data['month'].unique()):
                for day_name in list(data['day_name'].unique()):
                    df_aux = pd.DataFrame()
                    for day_type in list(data['day_type'].unique()):
                        data_filter = data[
                            (data['seasons'] == season) & (data['month'] == month) & (data['day_name'] == day_name) & (data['day_type'] == day_type)].copy()
                        if not data_filter.empty:
                            data_filter['sce'] = aux
                            df_aux = pd.concat([df_aux, data_filter])
                            aux += 1
                            scenarios[f'{season}_{month}_{day_name}_{day_type}'] = df_aux
        return scenarios

def get_mean(array):
    array = np.asarray(array)
    array = np.reshape(array, (-1, 24))
    array = np.mean(array, axis=0)
    return array


def get_std_dev(array):
    array = np.asarray(array)
    array = np.reshape(array, (-1, 24))
    array = np.std(array, axis=0)
    return array


# -*- coding: utf-8 -*-
# @Time    : 10/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

import pandas as pd
import os
import pathlib
import openpy_fxts.models.forecast as fxts


class run_all_models:

    def __init__(
            self,
            config_data: dict = None,
            config_mdl: dict = None,
            config_sim: dict = None,
            config_arch: dict = None,
            source: str = None,
            type_mdl: str = None,
            name_mdl: str = None,
            train: int = None,
            test: int = None,
            folder: str = 'architectures',
            view: bool = False
    ):
        self.config_data = config_data
        self.config_mdl = config_mdl
        self.config_sim = config_sim
        self.config_arch = config_arch
        self.source = source
        self.type_mdl = type_mdl
        self.name_mdl = name_mdl
        self.train = train
        self.test = test
        self.folder = folder
        self.view = view
        name_folder, mttr_upd = fxts.date_init_final(self.config_data)
        self.name_folder = name_folder
        self.mttr_upd = mttr_upd

    def _run_train(self, object_class):
        mdl = object_class(self.config_data, self.config_mdl, self.config_sim, self.config_arch)
        if mdl.name_mdl is None:
            mdl.name_mdl = self.name_mdl
        print(mdl.name_mdl)
        path = pathlib.Path(self.source).joinpath(
            self.folder, mdl.type_mdl, mdl.name_mdl, self.mttr_upd, self.name_folder
        )
        return mdl.training_mdl(path)

    def _run_test(self, object_class):
        mdl = object_class(self.config_data, self.config_mdl, self.config_sim, self.config_arch)
        if mdl.name_mdl is None:
            mdl.name_mdl = self.name_mdl
        print(mdl.name_mdl)
        path = pathlib.Path(self.source).joinpath(
            self.folder, mdl.type_mdl, mdl.name_mdl, self.mttr_upd, self.name_folder
        )
        return mdl.prediction_mdl(path)

    def run_case_mdl(self, object_class):
        if self.train == 1 and self.test == 0:
            return self._run_train(object_class)
        if self.train == 0 and self.test == 1:
            return self._run_test(object_class)
        if self.train == 1 and self.test == 1:
            return self._run_test(object_class)

    def execute_case(self):
        # Simple - Hybrids models_fxts
        if self.type_mdl == 'MLP':
            return self._mlp()
        if self.type_mdl == 'BiRNN':
            return self._birnn()
        if self.type_mdl == 'Conv1D':
            return self._conv1d()
        if self.type_mdl == 'Others':
            return self._others()
        if self.type_mdl == 'EncDec':
            return self._encdec()
        else:
            return None
        '''
        if self.type_mdl == 'Stacked':
            return self._stacked()
        if self.type_mdl == 'seq2seq':
            return self._seq2seq()
        '''


    def _mlp(self):
        if self.name_mdl == 'MLP_Dense':
            return self.run_case_mdl(fxts.MLP_Dense_class)

    def _birnn(self):
        # Simple models_fxts
        if self.name_mdl == 'BiRNN_Dense':
            return self.run_case_mdl(fxts.BiRNN_Dense_class)
        if self.name_mdl == 'Multi_BiRNN_Dense':
            return self.run_case_mdl(fxts.Multi_BiRNN_Dense_class)
        # Hybrids models_fxts
        if self.name_mdl == 'BiRNN_Conv1D':
            return self.run_case_mdl(fxts.BiRNN_Conv1D_class)
        if self.name_mdl == 'BiRNN_TimeDist_Dense':
            return self.run_case_mdl(fxts.BiRNN_TimeDist_Dense_class)
        if self.name_mdl == 'BiRNN_Bahdanau_Attention_Conv1D':
            return self.run_case_mdl(fxts.BiRNN_Bahdanau_Attention_Conv1D_class)
        if self.name_mdl == 'BiRNN_Luong_Attention_Conv1D':
            return self.run_case_mdl(fxts.BiRNN_Luong_Attention_Conv1D_class)
        if self.name_mdl == 'BiRNN_MultiHeadAttention_Conv1D':
            return self.run_case_mdl(fxts.BiRNN_MultiHeadAttention_Conv1D_class)
        if self.name_mdl == 'BiRNN_MDN':
            pass
            #return self.run_case_mdl(fxts.BiRNN_MDN)

    def _conv1d(self):
        # Simple
        if self.name_mdl == 'Conv1D_Dense':
            return self.run_case_mdl(fxts.Conv1D_Dense_class)
        if self.name_mdl == 'Multi_Conv1D_Dense':
            return self.run_case_mdl(fxts.Multi_Conv1D_Dense_class)
        # Hybrids
        if self.name_mdl == 'Conv1D_BiRNN':
            return self.run_case_mdl(fxts.Conv1D_BiRNN_class)
        if self.name_mdl == 'TimeDist_Conv1D_BiRNN':
            return self.run_case_mdl(fxts.TimeDist_Conv1D_BiRNN_class)
        if self.name_mdl == 'Conv1D_BiRNN_Attention':
            return self.run_case_mdl(fxts.Conv1D_BiRNN_Attention_class)

    def _others(self):
        if self.name_mdl == 'TCN_BiRNN':
            return self.run_case_mdl(fxts.TCN_BiRNN_class)
        if self.name_mdl == 'Time2Vec_BiRNN':
            return self.run_case_mdl(fxts.Time2Vec_BiRNN_class)

    def _encdec(self):
        if self.name_mdl == 'EncDec_BiRNN':
            return self.run_case_mdl(fxts.EncDec_BiRNN_class)
        if self.name_mdl == 'EncDec_Conv1D_BiRNN':
            return self.run_case_mdl(fxts.EncDec_Conv1D_BiRNN_class)
    '''
    def _stacked(self):
        if self.name_mdl == 'BiRNN_Stacked':
            return self.run_case_mdl(forecast.RNN_Stacked)
        if self.name_mdl == 'BiRNN_Stacked':
            return self.run_case_mdl(forecast.BiLSTM_Stacked)
        if self.name_mdl == 'Stacked_SciNet':
            return self.run_case_mdl(forecast.Stacked_SciNet)

    def _seq2seq(self):
        if self.name_mdl == 'Seq2Seq_RNN':
            return self.run_case_mdl(forecast.Seq2Seq_RNN_class)
        if self.name_mdl == 'Seq2Seq_RNN2':
            return self.run_case_mdl(forecast.Seq2Seq_RNN2_class)
        if self.name_mdl == 'Seq2Seq_RNN_Batch_Drop':
            return self.run_case_mdl(forecast.Seq2Seq_RNN_Batch_Drop_class)
        if self.name_mdl == 'Seq2Seq_BiRNN':
            return self.run_case_mdl(forecast.Seq2Seq_BiRNN_class)
        if self.name_mdl == 'Seq2Seq_BiLSTM2':
            return self.run_case_mdl(forecast.Seq2Seq_BiRNN2_class)
        if self.name_mdl == 'Seq2Seq_Conv1D_BiRNN':
            return self.run_case_mdl(forecast.Seq2Seq_Conv1D_BiRNN_class)
        if self.name_mdl == 'Seq2Seq_Multi_Head_Conv1D_BiRNN':
            return self.run_case_mdl(forecast.Seq2Seq_Multi_Head_Conv1D_BiRNN_class)
        if self.name_mdl == 'Seq2Seq_RNN_with_Luong_Attention':
            return self.run_case_mdl(forecast.Seq2Seq_RNN_with_Luong_Attention_class)
        if self.name_mdl == 'Seq2Seq_BiRNN_with_Attention':
            return self.run_case_mdl(forecast.Seq2Seq_BiRNN_with_Attention_class)
        if self.name_mdl == 'Seq2Seq_Conv2D_RNN':
            return self.run_case_mdl(forecast.Seq2Seq_Conv2D_RNN_class)
    '''


# -*- coding: utf-8 -*-
# @Time    : 10/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

from openpy_fxts.base_tf import tkm, tkl, tf
from openpy_fxts.architectures.mdits.arch_dicts import _get_dict_mdits
from openpy_fxts.architectures.layers.layers_class import dense_multi_out, dense_drop
from openpy_fxts.base_tf import value_miss


class mask_dense(tkl.Layer):
    def __init__(self, units, activation):
        super().__init__()
        self.mask = tkl.Masking(mask_value=value_miss)
        self.dense = tkl.Dense(units=units, activation=activation)

    def call(self, inputs):
        x = self.mask(inputs)
        x = self.dense(x)
        return x


class mlp_dense_class(tkm.Model):

    def __init__(
            self,
            n_past=None,
            n_inp_ft=None,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(mlp_dense_class, self).__init__()
        if config is None:
            config = _get_dict_mdits('MLP_Dense').config_mlp()
        self.n_past = n_past
        self.n_inp_ft = n_inp_ft
        self.n_future = n_future
        self.n_out_ft = n_out_ft
        self.config = config
        self.hidden_input = []
        for _ in range(n_inp_ft):
            self.hidden_input.append(
                mask_dense(
                    units=config['input_layer']['units'],
                    activation=config['input_layer']['activation']
                )
            )
        self.hidden_dense = dense_drop(
            n_units=config['dense_layers']['units'],
            list_act=config['dense_layers']['activations'],
            use_drop=config['dropout']['activate'],
            rate=config['dropout']['rate']
        )
        self.dense_out = dense_multi_out(n_future, n_out_ft, config['output_layer']['activation'])

    def call(self, inputs, training=True, **kwargs):
        x_input = []
        for i, (layer_input) in enumerate(self.hidden_input):
            x_input.append(layer_input(inputs[i]))
        x = tkl.concatenate(x_input)
        x = self.hidden_dense(x)
        x = self.dense_out(x)
        return x




class _dense_drop_block(tkl.Layer):
    def __init__(self, units, activation, regularizer, dropout):
        super().__init__()
        self.dense = tkl.Dense(units=units, activation=activation, kernel_regularizer=tkr.L2(regularizer))
        self.dropout = tkl.Dropout(dropout)

    def call(self, inputs):
        x = self.dense(inputs)
        x = self.dropout(x)
        return x


class mlp_dense(tf.keras.Model):
    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(mlp_dense, self).__init__()

        if config is None:
            config = {
                'layer_config': {
                    'hidden_dimensions': [50, 25, 25],
                    'hidden_activations': ['relu', 'relu', 'relu'],
                    'hidden_regularization': [None, None, None],
                },
                'dropout': 0.2,
                'layer_out': 'relu'
            }
        hidden_dimensions = config['layer_config']['hidden_dimensions']
        hidden_activations = config['layer_config']['hidden_activations']
        hidden_regularization = config['layer_config']['hidden_regularization']

        self.mask = tkl.Masking(mask_value=value_miss)
        self.hidden_layers = dict()
        for i, (unit, act, reg) in enumerate(zip(hidden_dimensions, hidden_activations, hidden_regularization)):
            self.hidden_layers['hidden_' + str(i)] = _dense_drop_block(unit, act, reg, config['dropout'])
        self.dense_out = tkl.Dense(n_future * n_out_ft, activation=config['layer_out'])

    def call(self, inputs, training=True, **kwargs):
        x = self.mask(inputs)
        for _, layer in self.hidden_layers.items():
            x = layer(x)
        x = self.dense_out(x)
        return x



from openpy_fxts.models.clustering.prepare_data import get_scenarios_for_time_features, get_mean, get_std_dev
from openpy_fxts.models.clustering.project_app import Mult_escenarios

# New files
from openpy_fxts.models.clustering.data_preprocessing import read_BBDD
from openpy_fxts.models.clustering.data_preprocessing import scenarios
read_BBDD = read_BBDD()
scenarios = scenarios()

from openpy_fxts.models.clustering.clustering_algorithms import clustering_ts

# Utils plot
from openpy_fxts.models.clustering.utils_plot import plot_scenarios





from openpy_fxts.baseline_mdls import base_class


class Conv1D_Dense_class(base_class):

    def __init__(
            self,
            config_data=None,
            config_mdl=None,
            config_sim=None,
            config_arch=None,
            name_mdl='Conv1D_Dense',
            type_mdl='Conv1D',
            app='mdits'
    ):
        super().__init__(config_data, config_mdl, config_sim, config_arch, name_mdl, type_mdl, app)
        return


class Multi_Conv1D_Dense_class(base_class):

    def __init__(
            self,
            config_data=None,
            config_mdl=None,
            config_sim=None,
            config_arch=None,
            name_mdl='Multi_Conv1D_Dense',
            type_mdl='Conv1D',
            app='mdits'
    ):
        super().__init__(config_data, config_mdl, config_sim, config_arch, name_mdl, type_mdl, app)
        return



from openpy_fxts.mdits.mdls_dl.base_tf import tkl, tfk, K


class attention(tkl.Layer):

    def __init__(self, **kwargs):
        super(attention, self).__init__(**kwargs)

    def build(self, input_shape):
        self.W = self.add_weight(
            name='attention_weight',
            shape=(input_shape[-1], 1),
            initializer='random_normal',
            trainable=True
        )
        self.b = self.add_weight(
            name='attention_bias',
            shape=(input_shape[1], 1),
            initializer='zeros',
            trainable=True
        )
        super(attention, self).build(input_shape)

    def call(self, x):
        # Alignment scores. Pass them through tanh function
        e = K.tanh(K.dot(x, self.W) + self.b)
        # Remove dimension of size 1
        e = K.squeeze(e, axis=-1)
        # Compute the weights
        alpha = K.softmax(e)
        # Reshape to tensorFlow format
        alpha = K.expand_dims(alpha, axis=-1)
        # Compute the context vector
        context = x * alpha
        context = K.sum(context, axis=1)
        return context


class SeqSelfAttention(tkl.Layer):

    ATTENTION_TYPE_ADD = 'additive'
    ATTENTION_TYPE_MUL = 'multiplicative'

    def __init__(
            self,
            units=32,
            attention_width=None,
            attention_type=ATTENTION_TYPE_ADD,
            return_attention=False,
            history_only=False,
            kernel_initializer='glorot_normal',
            bias_initializer='zeros',
            kernel_regularizer=None,
            bias_regularizer=None,
            kernel_constraint=None,
            bias_constraint=None,
            use_additive_bias=True,
            use_attention_bias=True,
            attention_activation=None,
            attention_regularizer_weight=0.0,
            **kwargs
    ):
        """Layer initialization.
        For additive attention, see: https://arxiv.org/pdf/1806.01264.pdf
        :param units: The dimension of the vectors that used to calculate the attention weights.
        :param attention_width: The width of local attention.
        :param attention_type: 'additive' or 'multiplicative'.
        :param return_attention: Whether to return the attention weights for visualization.
        :param history_only: Only use historical pieces of datasets.
        :param kernel_initializer: The initializer for weight matrices.
        :param bias_initializer: The initializer for biases.
        :param kernel_regularizer: The regularization for weight matrices.
        :param bias_regularizer: The regularization for biases.
        :param kernel_constraint: The constraint for weight matrices.
        :param bias_constraint: The constraint for biases.
        :param use_additive_bias: Whether to use bias while calculating the relevance of inputs features in additive mode.
        :param use_attention_bias: Whether to use bias while calculating the weights of attention.
        :param attention_activation: The activation used for calculating the weights of attention.
        :param attention_regularizer_weight: The weights of attention regularizer.
        :param kwargs: Parameters for parent class.
        """
        super(SeqSelfAttention, self).__init__(**kwargs)
        self.supports_masking = True
        self.units = units
        self.attention_width = attention_width
        self.attention_type = attention_type
        self.return_attention = return_attention
        self.history_only = history_only
        if history_only and attention_width is None:
            self.attention_width = int(1e9)

        self.use_additive_bias = use_additive_bias
        self.use_attention_bias = use_attention_bias
        self.kernel_initializer = tfk.initializers.get(kernel_initializer)
        self.bias_initializer = tfk.initializers.get(bias_initializer)
        self.kernel_regularizer = tfk.regularizers.get(kernel_regularizer)
        self.bias_regularizer = tfk.regularizers.get(bias_regularizer)
        self.kernel_constraint = tfk.constraints.get(kernel_constraint)
        self.bias_constraint = tfk.constraints.get(bias_constraint)
        self.attention_activation = tfk.activations.get(attention_activation)
        self.attention_regularizer_weight = attention_regularizer_weight
        self._backend = K.backend()

        if attention_type == SeqSelfAttention.ATTENTION_TYPE_ADD:
            self.Wx, self.Wt, self.bh = None, None, None
            self.Wa, self.ba = None, None
        elif attention_type == SeqSelfAttention.ATTENTION_TYPE_MUL:
            self.Wa, self.ba = None, None
        else:
            raise NotImplementedError('No implementation for attention type : ' + attention_type)

    def get_config(self):
        config = {
            'units': self.units,
            'attention_width': self.attention_width,
            'attention_type': self.attention_type,
            'return_attention': self.return_attention,
            'history_only': self.history_only,
            'use_additive_bias': self.use_additive_bias,
            'use_attention_bias': self.use_attention_bias,
            'kernel_initializer': tfk.initializers.serialize(self.kernel_initializer),
            'bias_initializer': tfk.initializers.serialize(self.bias_initializer),
            'kernel_regularizer': tfk.regularizers.serialize(self.kernel_regularizer),
            'bias_regularizer': tfk.regularizers.serialize(self.bias_regularizer),
            'kernel_constraint': tfk.constraints.serialize(self.kernel_constraint),
            'bias_constraint': tfk.constraints.serialize(self.bias_constraint),
            'attention_activation': tfk.activations.serialize(self.attention_activation),
            'attention_regularizer_weight': self.attention_regularizer_weight,
        }
        base_config = super(SeqSelfAttention, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def build(self, input_shape):
        if self.attention_type == SeqSelfAttention.ATTENTION_TYPE_ADD:
            self._build_additive_attention(input_shape)
        elif self.attention_type == SeqSelfAttention.ATTENTION_TYPE_MUL:
            self._build_multiplicative_attention(input_shape)
        super(SeqSelfAttention, self).build(input_shape)

    def _build_additive_attention(self, input_shape):
        feature_dim = int(input_shape[2])

        self.Wt = self.add_weight(
            shape=(feature_dim, self.units),
            name='{}_Add_Wt'.format(self.name),
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint
        )
        self.Wx = self.add_weight(
            shape=(feature_dim, self.units),
            name='{}_Add_Wx'.format(self.name),
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint
        )
        if self.use_additive_bias:
            self.bh = self.add_weight(
                shape=(self.units,),
                name='{}_Add_bh'.format(self.name),
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint
            )
        self.Wa = self.add_weight(
            shape=(self.units, 1),
            name='{}_Add_Wa'.format(self.name),
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint
        )
        if self.use_attention_bias:
            self.ba = self.add_weight(
                shape=(1,),
                name='{}_Add_ba'.format(self.name),
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint
            )

    def _build_multiplicative_attention(self, input_shape):
        feature_dim = int(input_shape[2])

        self.Wa = self.add_weight(
            shape=(feature_dim, feature_dim),
            name='{}_Mul_Wa'.format(self.name),
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint
        )
        if self.use_attention_bias:
            self.ba = self.add_weight(
                shape=(1,),
                name='{}_Mul_ba'.format(self.name),
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint
            )

    def call(self, inputs, mask=None, **kwargs):
        input_len = K.shape(inputs)[1]

        if self.attention_type == SeqSelfAttention.ATTENTION_TYPE_ADD:
            e = self._call_additive_emission(inputs)
        elif self.attention_type == SeqSelfAttention.ATTENTION_TYPE_MUL:
            e = self._call_multiplicative_emission(inputs)

        if self.attention_activation is not None:
            e = self.attention_activation(e)
        if self.attention_width is not None:
            if self.history_only:
                lower = K.arange(0, input_len) - (self.attention_width - 1)
            else:
                lower = K.arange(0, input_len) - self.attention_width // 2
            lower = K.expand_dims(lower, axis=-1)
            upper = lower + self.attention_width
            indices = K.expand_dims(K.arange(0, input_len), axis=0)
            e -= 10000.0 * (1.0 - K.cast(lower <= indices, K.floatx()) * K.cast(indices < upper, K.floatx()))
        if mask is not None:
            mask = K.expand_dims(K.cast(mask, K.floatx()), axis=-1)
            e -= 10000.0 * ((1.0 - mask) * (1.0 - K.permute_dimensions(mask, (0, 2, 1))))

        # a_{t} = \text{softmax}(e_t)
        e = K.exp(e - K.max(e, axis=-1, keepdims=True))
        a = e / K.sum(e, axis=-1, keepdims=True)

        # l_t = \sum_{t'} a_{t, t'} x_{t'}
        v = K.batch_dot(a, inputs)
        if self.attention_regularizer_weight > 0.0:
            self.add_loss(self._attention_regularizer(a))

        if self.return_attention:
            return [v, a]
        return v

    def _call_additive_emission(self, inputs):
        input_shape = K.shape(inputs)
        batch_size, input_len = input_shape[0], input_shape[1]

        # h_{t, t'} = \tanh(x_t^T W_t + x_{t'}^T W_x + b_h)
        q = K.expand_dims(K.dot(inputs, self.Wt), 2)
        k = K.expand_dims(K.dot(inputs, self.Wx), 1)
        if self.use_additive_bias:
            h = K.tanh(q + k + self.bh)
        else:
            h = K.tanh(q + k)

        # e_{t, t'} = W_a h_{t, t'} + b_a
        if self.use_attention_bias:
            e = K.reshape(K.dot(h, self.Wa) + self.ba, (batch_size, input_len, input_len))
        else:
            e = K.reshape(K.dot(h, self.Wa), (batch_size, input_len, input_len))
        return e

    def _call_multiplicative_emission(self, inputs):
        # e_{t, t'} = x_t^T W_a x_{t'} + b_a
        e = K.batch_dot(K.dot(inputs, self.Wa), K.permute_dimensions(inputs, (0, 2, 1)))
        if self.use_attention_bias:
            e += self.ba[0]
        return e

    def compute_output_shape(self, input_shape):
        output_shape = input_shape
        if self.return_attention:
            attention_shape = (input_shape[0], output_shape[1], input_shape[1])
            return [output_shape, attention_shape]
        return output_shape

    def compute_mask(self, inputs, mask=None):
        if self.return_attention:
            return [mask, None]
        return mask

    def _attention_regularizer(self, attention):
        batch_size = K.cast(K.shape(attention)[0], K.floatx())
        input_len = K.shape(attention)[-1]
        indices = K.expand_dims(K.arange(0, input_len), axis=0)
        diagonal = K.expand_dims(K.arange(0, input_len), axis=-1)
        eye = K.cast(K.equal(indices, diagonal), K.floatx())
        return self.attention_regularizer_weight * K.sum(K.square(K.batch_dot(
            attention,
            K.permute_dimensions(attention, (0, 2, 1))) - eye)) / batch_size

    @staticmethod
    def get_custom_objects():
        return {'SeqSelfAttention': SeqSelfAttention}



# -*- coding: utf-8 -*-
# @Time    : 10/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

from openpy_fxts.base_tf import tkm, tkl, value_miss
from openpy_fxts.architectures.layers.layers_class import StackedSciNet


class lstm_stacked(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(lstm_stacked, self).__init__()
        if config is None:
            config = {
                'bilstm': {
                    'units': 256,
                    'activation': 'tanh',
                    'recurrent_activation': 'relu'
                },
                'dropout': 0.3,
                'dense': {
                    'activation': 'linear'
                }
            }
        self.mask = tkl.Masking(mask_value=value_miss)
        self.lstm1 = tkl.LSTM(units=256, activation='relu', return_sequences=True)
        self.lstm2 = tkl.LSTM(units=256, activation='relu')
        self.dense1 = tkl.Dense(units=128)
        self.dropout = tkl.Dropout(0.3)
        self.dense2 = tkl.Dense(n_future * n_out_ft, activation='relu')
        self.reshape = tkl.Reshape((n_future, n_out_ft))
        self.conv1d = tkl.Conv1D(n_out_ft, 1, padding='same')

    def call(self, inputs, training=True, **kwargs):
        x = self.mask(inputs)
        x = self.lstm1(x)
        x = self.lstm2(x)
        x = self.dense1(x)
        x = self.dropout(x)
        x = self.dense2(x)
        x = self.reshape(x)
        x = self.conv1d(x)
        return x

class bilstm_stacked(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(bilstm_stacked, self).__init__()
        if config is None:
            config = {
                'bilstm': {
                    'units': 256,
                    'activation': 'tanh',
                    'recurrent_activation': 'relu'
                },
                'dropout': 0.3,
                'dense': {
                    'activation': 'linear'
                }
            }
        self.mask = tkl.Masking(mask_value=value_miss)
        self.bilstm1 = tkl.Bidirectional(tkl.LSTM(units=128, activation='relu', return_sequences=True))
        self.bilstm2 = tkl.Bidirectional(tkl.LSTM(units=128, activation='relu'))
        self.dense1 = tkl.Dense(units=128)
        self.dropout = tkl.Dropout(0.3)
        self.dense2 = tkl.Dense(n_future * n_out_ft, activation='relu')
        self.reshape = tkl.Reshape((n_future, n_out_ft))
        self.conv1d = tkl.Conv1D(n_out_ft, 1, padding='same')

    def call(self, inputs, training=True, **kwargs):
        x = self.mask(inputs)
        x = self.bilstm1(x)
        x = self.bilstm2(x)
        x = self.dense1(x)
        x = self.dropout(x)
        x = self.dense2(x)
        x = self.reshape(x)
        x = self.conv1d(x)
        return x


class stackedscinet(tkm.Model):
    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(stackedscinet, self).__init__()
        if config is None:
            config = {
                'bilstm': {
                    'units': 256,
                    'activation': 'tanh',
                    'recurrent_activation': 'relu'
                },
                'dropout': 0.3,
                'dense': {
                    'activation': 'linear'
                }
            }
        self.mask = tkl.Masking(mask_value=value_miss)
        self.stackedscinet = StackedSciNet(
            horizon=n_future,
            features=n_out_ft,
            stacks=2,
            levels=4,
            h=4,
            kernel_size=5
        )

    def call(self, inputs, training=True, **kwargs):
        x = self.mask(inputs)
        x = self.stackedscinet(x)
        return x





# -*- coding: utf-8 -*-
# @Time    : 21/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

from openpy_fxts.base_tf import tkm, tkl
from openpy_fxts.architectures.mdits.arch_dicts import _get_dict_mdits
from openpy_fxts.architectures.layers.layers_class import type_rnn, multi_rnn_drop, multi_conv1d_pool_flat_drop


class encdec_birnn(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(encdec_birnn, self).__init__()
        if config is None:
            config = _get_dict_mdits('EncDec_BiRNN').config_encdec()
        self.n_future = n_future,
        self.n_out_ft = n_out_ft,
        self.config = config
        self.encoder = multi_rnn_drop(
            mdl=type_rnn(config['encoder']['rnn_layers']['type']),
            units=config['encoder']['rnn_layers']['units'],
            activations=config['encoder']['rnn_layers']['activations'],
            sequences=config['encoder']['rnn_layers']['sequences'],
            bidirectional=config['encoder']['rnn_layers']['bidirectional'],
            use_drop=config['dropout']['activate'],
            rate=config['dropout']['rate'],
            is_mask=True
        )
        self.rep_vec = tkl.RepeatVector(n_future)
        self.decoder = multi_rnn_drop(
            mdl=type_rnn(config['decoder']['rnn_layers']['type']),
            units=config['decoder']['rnn_layers']['units'],
            activations=config['decoder']['rnn_layers']['activations'],
            sequences=config['decoder']['rnn_layers']['sequences'],
            bidirectional=config['decoder']['rnn_layers']['bidirectional'],
            use_drop=config['dropout']['activate'],
            rate=config['dropout']['rate'])
        self.time_dense = tkl.TimeDistributed(tkl.Dense(n_out_ft))

    def call(self, inputs):
        encdec = self.encoder(inputs)
        encdec = self.rep_vec(encdec)
        encdec = self.decoder(encdec)
        encdec = self.time_dense(encdec)
        return encdec


class encdec_conv1d_birnn(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(encdec_conv1d_birnn, self).__init__()
        if config is None:
            config = _get_dict_mdits('EncDec_Conv1D_BiRNN').config_encdec()
        self.n_future = n_future,
        self.n_out_ft = n_out_ft,
        self.config = config
        self.encoder = multi_conv1d_pool_flat_drop(
            filters=config['encoder']['conv1d_layers']['filters'],
            kernels=config['encoder']['conv1d_layers']['kernels'],
            activations=config['encoder']['conv1d_layers']['activations'],
            paddings=config['encoder']['conv1d_layers']['paddings'],
            use_drop=config['dropout']['activate'],
            rate=config['dropout']['rate'],
            maxpooling_1d=True,
            pool_size=config['encoder']['conv1d_layers']['pool_size'],
            flatten=True,
            is_mask=True
        )
        self.rep_vec = tkl.RepeatVector(n_future)
        self.decoder = multi_rnn_drop(
            mdl=type_rnn(config['decoder']['rnn_layers']['type']),
            units=config['decoder']['rnn_layers']['units'],
            activations=config['decoder']['rnn_layers']['activations'],
            sequences=config['decoder']['rnn_layers']['sequences'],
            bidirectional=config['decoder']['rnn_layers']['bidirectional'],
            use_drop=config['dropout']['activate'],
            rate=config['dropout']['rate']
        )
        self.time_dense = tkl.TimeDistributed(tkl.Dense(n_out_ft))

    def call(self, inputs):
        encdec = self.encoder(inputs)
        encdec = self.rep_vec(encdec)
        encdec = self.decoder(encdec)
        encdec = self.time_dense(encdec)
        return encdec


# -*- coding: utf-8 -*-
# @Time    : 31/08/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

import pandas as pd
import matplotlib.pyplot as plt
from datetime import time


def get_curves_typical(
        data: pd.DataFrame,
        colums_load: list,
        name_sce: str,
        plot: bool = False,
        path_save: str = None,
        view: bool = True
):
    data['aux2'] = data.hour.astype(str) + ':' + data.minute.astype(str)
    data['mean'] = data[colums_load].mean(axis=1)
    data['std'] = data[colums_load].std(axis=1)
    data['aux'] = data.index
    data['aux'] = data.aux.dt.strftime('%H:%M')
    df_curve_typical = data.pivot_table(index=['hour', 'minute'], values=['mean', 'std'])
    df_curve_typical['aux2'] = data.index.strftime('%H:%M').unique()
    df_curve_typical.set_index('aux2', inplace=True)
    if plot:
        df_AllUsers = data.pivot_table(index=['hour', 'minute'], values=colums_load)
        df_AllUsers['aux2'] = data.index.strftime('%H:%M').unique()
        df_AllUsers.set_index('aux2', inplace=True)
        df_summary = pd.concat([df_curve_typical, df_AllUsers], axis=1)
        df_summary['aux2'] = data.index.strftime('%H:%M').unique()
        df_summary.set_index('aux2', inplace=True)
        fig = plt.figure()
        plt.figure().clear()
        for col in df_summary.columns:
            if col == 'mean':
                plt.plot(
                    df_summary[col],
                    "r-",
                    label='mean'
                )
            if col == 'std':
                plt.plot(
                    df_summary[col],
                    "g-",
                    label='std'
                )
            else:
                plt.plot(
                    df_summary[col],
                    "k-",
                    alpha=.15,
                    #label=str(col)
                )
        plt.xticks(rotation=90)
        plt.legend(loc='lower right')
        plt.title(name_sce)
        if view:
            plt.show()
        if path_save is not None:
            plt.savefig(f'{path_save}\\{name_sce}.png')
    return df_curve_typical





# -*- coding: utf-8 -*-
# @Time    : 10/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

from openpy_fxts.preprocessing_data.forecast.prepare_data import pre_processing_data
from openpy_fxts.models.utils import _callbacks, _learning_curve
from openpy_fxts.models.utils import _process_values_preliminary
from openpy_fxts.models.utils import _mdl_caracteristic
import tensorflow as tf

tkm = tf.keras.models
tkl = tf.keras.layers
tkloss = tf.keras.losses
tko = tf.keras.optimizers
tku = tf.keras.utils


class Multi_Layer_Perceptron:

    def __init__(self, config_data=None, config_mdl=None, config_sim=None):
        # Parameters for dataset
        self.config_data = config_data
        self.n_past = config_data['n_past']
        self.n_future = config_data['n_future']
        self.n_inp_ft = config_data['n_inp_ft']
        self.n_out_ft = config_data['n_out_ft']
        # Parameters for neural network
        self.config_mdl = config_mdl
        self.optimizer = config_mdl['optimizer']
        self.loss = config_mdl['loss']
        self.metrics = config_mdl['metrics']
        self.batch_size = config_mdl['batch_size']  # Batch size for training.
        self.epochs = config_mdl['epochs']  # Number of epochs to train for.
        self.units = config_mdl['units']  # no of lstm units
        self.dropout = config_mdl['dropout']
        # Parameters for simulation
        self.config_sim = config_sim
        self.verbose = config_sim['verbose']
        self.patience = config_sim['patience']
        self.plot_history = config_sim['plt_history']
        self.preliminary = config_sim['preliminary']
        self.name_model = 'mlp'

    def build_model(self):
        n_input = self.n_past * self.n_inp_ft
        n_output = self.n_future * self.n_out_ft
        # Define model.
        model = tkm.Sequential()
        model.add(tkl.Dense(self.units, activation='relu', input_dim=n_input))
        model.add(tkl.Dense(n_output))
        return model

    def train_model(
            self,
            filepath: str = None,
    ):
        data_test = pre_processing_data(self.config_data, train=True, valid=True)
        pre_processed = data_test.transformer_data()
        model = Multi_Layer_Perceptron(self.config_data, self.config_mdl).build_model()
        model.compile(
            optimizer=self.optimizer,
            loss=self.loss,
            metrics=self.metrics
        )
        _mdl_caracteristic(model, self.config_sim, filepath)
        # Training
        X_train = pre_processed['train']['X']
        y_train = pre_processed['train']['y']
        X_train = X_train.reshape(X_train.shape[0], self.n_past * self.n_inp_ft)
        y_train = y_train.reshape(y_train.shape[0], self.n_future * self.n_out_ft)
        # Validation
        X_valid = pre_processed['valid']['X']
        y_valid = pre_processed['valid']['y']
        X_valid = X_valid.reshape(X_valid.shape[0], self.n_past * self.n_inp_ft)
        y_valid = y_valid.reshape(y_valid.shape[0], self.n_future * self.n_out_ft)

        history = model.fit(
            X_train,
            y_train,
            epochs=self.epochs,
            validation_data=(
                X_valid,
                y_valid
            ),
            batch_size=self.batch_size,
            verbose=self.verbose,
            callbacks=_callbacks(
                filepath,
                weights=True,
                verbose=self.verbose,
                patience=self.patience
            )
        )
        if self.config_data['plt_history']:
            _learning_curve(history, self.name_model, filepath, self.config_sim['time_init'])
        if self.config_data['preliminary']:
            data_test = pre_processing_data(self.config_data, test=True)
            dict_test = data_test.transformer_data()
            yhat = _process_values_preliminary(
                model,
                dict_test,
                self.config_data,
                self.config_sim,
                self.name_model
            ).get_values(type_mdl='Dense')
        return model

    def prediction(
            self,
            model
    ):
        data = pre_processing_data(self.config_data, test=True)
        dict_test = data.transformer_data()
        yhat =_process_values_preliminary(
            model,
            dict_test,
            self.config_data,
            self.config_sim,
            self.name_model
        ).get_values(type_mdl='Dense')
        return yhat


# -*- coding: utf-8 -*-
# @Time    : 06/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm


import pandas as pd
import numpy as np
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import MinMaxScaler

from openpy_fxts.preprocessing_data.imputation.utils_data import sin_transformer, cos_transformer, _filter_y_features, \
    _series_to_supervised, _scaler_data

pd.options.mode.chained_assignment = None


class feature_engineering:

    def __init__(self, df: pd.DataFrame):
        self.df = df

    def add_season(self, ):
        return self.df

    def add_temporaly(self, ):
        return self.df

    def trigonometric_features(self, categorical_columns: list[str] = None):
        one_hot_encoder = OneHotEncoder(handle_unknown="ignore", sparse_output=False)
        cyclic_cossin_transformer = ColumnTransformer(
            transformers=[
                ("categorical", one_hot_encoder, categorical_columns),
                ("month_sin", sin_transformer(12), ["month"]),
                ("month_cos", cos_transformer(12), ["month"]),
                ("weekday_sin", sin_transformer(7), ["weekday"]),
                ("weekday_cos", cos_transformer(7), ["weekday"]),
                ("hour_sin", sin_transformer(24), ["hour"]),
                ("hour_cos", cos_transformer(24), ["hour"]),
            ],
            remainder=MinMaxScaler(),
        )
        return cyclic_cossin_transformer

    def periodic_spline_features(self, categorical_columns):
        from sklearn.preprocessing import SplineTransformer
        one_hot_encoder = OneHotEncoder(handle_unknown="ignore", sparse_output=False)

        def periodic_spline_transformer(period, n_splines=None, degree=3):
            if n_splines is None:
                n_splines = period
            n_knots = n_splines + 1  # periodic and include_bias is True
            return SplineTransformer(
                degree=degree,
                n_knots=n_knots,
                knots=np.linspace(0, period, n_knots).reshape(n_knots, 1),
                extrapolation="periodic",
                include_bias=True,
            )

        cyclic_spline_transformer = ColumnTransformer(
            transformers=[
                ("categorical", one_hot_encoder, categorical_columns),
                ("cyclic_month", periodic_spline_transformer(12, n_splines=6), ["month"]),
                ("cyclic_weekday", periodic_spline_transformer(7, n_splines=3), ["weekday"]),
                ("cyclic_hour", periodic_spline_transformer(24, n_splines=12), ["hour"]),
            ],
            remainder=MinMaxScaler(),
        )
        return cyclic_spline_transformer


class pre_processing_data:

    def __init__(
            self,
            config: dict = None,
            train: bool = None,
            valid: bool = None,
            test: bool = None,
    ):

        self.config = config
        self.train = train
        self.valid = valid
        self.test = test

        self.pct_valid = config['pct_valid']
        self.pct_test = config['pct_test']
        self.dataset = config['dataset']

    def split_ts_Train_Valid_Test(
            self
    ):
        if self.pct_valid is not None and self.pct_test is not None:
            p_train = 1 - self.pct_valid - self.pct_test
            n_train = int(self.dataset.shape[0] * p_train)
            n_valid = int(self.dataset.shape[0] * self.pct_valid)
            n_test = n_train + n_valid
            if self.train and self.valid:
                df_train = self.dataset.iloc[:n_train, :]
                df_valid = self.dataset.iloc[n_train: n_train + n_valid, :]
                print(f'train: {df_train.shape} - valid: {df_valid.shape}')
                return df_train, df_valid, None
            if self.test:
                df_test = self.dataset.iloc[n_test:, :]
                print(f'test: {df_test.shape}')
                return None, None, df_test
        if self.pct_valid is None and self.pct_test is not None:
            self.pct_valid = 0
            p_train = 1 - self.pct_valid - self.pct_test
            n_train = int(self.dataset.shape[0] * p_train)
            n_valid = int(self.dataset.shape[0] * self.pct_valid)
            n_test = n_train + n_valid
            if self.train and self.valid:
                df_train = self.dataset.iloc[:n_train, :]
                df_valid = self.dataset.iloc[n_train: n_train + n_valid, :]
                print(f'train: {df_train.shape} - valid: {df_valid.shape}')
                return df_train, None, None
            if self.test:
                df_test = self.dataset.iloc[n_test:, :]
                print(f'test: {df_test.shape}')
                return None, None, df_test
        if self.pct_valid is not None and self.pct_test is None:
            self.pct_test = 0
            p_train = 1 - self.pct_valid - self.pct_test
            n_train = int(self.dataset.shape[0] * p_train)
            n_valid = int(self.dataset.shape[0] * self.pct_valid)
            n_test = n_train + n_valid
            if self.train and self.valid:
                df_train = self.dataset.iloc[:n_train, :]
                df_valid = self.dataset.iloc[n_train: n_train + n_valid, :]
                print(f'train: {df_train.shape} - valid: {df_valid.shape}')
                return df_train, df_valid, None
            if self.test:
                df_test = self.dataset.iloc[n_test:, :]
                print(f'test: {df_test.shape}')
                return None, None, df_test
        else:
            if self.pct_test is None:
                self.pct_test = 0
            p_train = 1 - self.pct_test
            n_train = int(self.dataset.shape[0] * p_train)
            if self.train:
                df_train = self.dataset.iloc[:n_train, :]
                print(f'train: {df_train.shape}')
                return df_train, None, None
            else:
                df_test = self.dataset.iloc[n_train:, :]
                print(f'test: {df_test.shape}')
                return None, None, df_test

    def _pre_processed_data_pipeline(
            self,
            df_data=None,
            seasonal_features=True,
            diff_trend=True,
            exog=True,
            fourier_terms=True,
            dropnan=None
    ):

        df_scaler, scaler = _scaler_data(df_data)
        ###CALCULATE SEASONAL MEANS
        if seasonal_features:
            pass

        if diff_trend:
            pass

        if exog:
            pass

        if fourier_terms:
            pass

        ts_sequence = _series_to_supervised(
            data=df_scaler,
            dropnan=dropnan,
            n_in=self.config['n_past'],
            n_out=self.config['n_future'],
            feat_str_at_end=True,
            feat_lag_str='IP',
            feat_lead_str='OP'
        )
        # cambiar el monento de incluir al agregar informacion de la fecha y transformaciones.
        n_past = self.config['n_past'] * len(self.config['x_colname'])
        n_future = self.config['n_future'] * len(self.config['x_colname'])
        df_X, df_y = ts_sequence.iloc[:, :n_past], ts_sequence.iloc[:, -n_future:]
        df_y = _filter_y_features(df_y, self.config['y_colname'], self.config['n_future'])
        print(df_X.shape, df_y.shape)

        return df_X, df_y, scaler

    def transformer_data(
            self,
            view: bool = True,
            dropnan: bool = None
    ):
        if view:
            print(f'Dataset -> {self.dataset.shape}')
        class_data = pre_processing_data(self.config, self.train, self.valid, self.test)
        df_train, df_validation, df_test, = class_data.split_ts_Train_Valid_Test()
        ts_pre_process = {}
        if self.train and self.valid:
            if self.pct_valid is not None:
                dict_train, dict_valid = {}, {}
                train_X, train_y, scaler_train = class_data._pre_processed_data_pipeline(
                    df_data=df_train,
                    dropnan=dropnan
                )
                valid_X, valid_y, scaler_valid = class_data._pre_processed_data_pipeline(
                    df_data=df_validation,
                    dropnan=dropnan
                )
                # reshape input to be 3D [samples, timesteps, features]
                # Training
                train_X = train_X.values.reshape(train_X.shape[0], self.config['n_past'], self.config['n_inp_ft'])
                train_y = train_y.values.reshape(train_y.shape[0], self.config['n_future'], self.config['n_out_ft'])
                # validation
                valid_X = valid_X.values.reshape(valid_X.shape[0], self.config['n_past'], self.config['n_inp_ft'])
                valid_y = valid_y.values.reshape(valid_y.shape[0], self.config['n_future'], self.config['n_out_ft'])
                if view:
                    print(f'Train -> X: {train_X.shape} - y:{train_y.shape}')
                    print(f'Valid -> X: {valid_X.shape} - y:{valid_y.shape}')
                dict_train['X'], dict_train['y'], dict_train['scaler'] = train_X, train_y, scaler_train
                dict_valid['X'], dict_valid['y'], dict_valid['scaler'] = valid_X, valid_y, scaler_valid
                ts_pre_process['train'], ts_pre_process['valid'] = dict_train, dict_valid
                return ts_pre_process
            if self.pct_valid is None:
                dict_train = {}
                train_X, train_y, scaler_train = class_data._pre_processed_data_pipeline(
                    df_data=df_train,
                    dropnan=dropnan
                )
                # reshape input to be 3D [samples, timesteps, features]
                # Training
                train_X = train_X.values.reshape(train_X.shape[0], self.config['n_past'], self.config['n_inp_ft'])
                train_y = train_y.values.reshape(train_y.shape[0], self.config['n_future'], self.config['n_out_ft'])
                if view:
                    print(f'Train -> X: {train_X.shape} - y:{train_y.shape}')
                dict_train['X'], dict_train['y'], dict_train['scaler'] = train_X, train_y, scaler_train
                ts_pre_process['train'], ts_pre_process['valid'] = dict_train
                return ts_pre_process
        elif self.train:
            dict_train = {}
            train_X, train_y, scaler_train = class_data._pre_processed_data_pipeline(
                df_data=df_train,
                dropnan=dropnan
            )
            # reshape input to be 3D [samples, timesteps, features]
            # Training
            train_X = train_X.values.reshape(train_X.shape[0], self.config['n_past'], self.config['n_inp_ft'])
            train_y = train_y.values.reshape(train_y.shape[0], self.config['n_future'], self.config['n_out_ft'])
            if view:
                print(f'Train -> X: {train_X.shape} - y:{train_y.shape}')

            dict_train['X'], dict_train['y'], dict_train['scaler'] = train_X, train_y, scaler_train
            return dict_train
        else:
            dict_test = {}
            test_X, test_y, scaler_test = class_data._pre_processed_data_pipeline(
                df_data=df_test,
                dropnan=dropnan
            )
            # Testing
            test_X = test_X.values.reshape(test_X.shape[0], self.config['n_past'], self.config['n_inp_ft'])
            test_y = test_y.values.reshape(test_y.shape[0], self.config['n_future'], self.config['n_out_ft'])
            if view:
                print(f'Test -> X: {test_X.shape} - y:{test_y.shape}')
            dict_test['X'], dict_test['y'], dict_test['scaler'] = test_X, test_y, scaler_test
            return dict_test



# -*- coding: utf-8 -*-
# @Time    : 10/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

from openpy_fxts.base_tf import tkm, tkl, value_miss
from openpy_fxts.architectures.layers.attention_class import attention


class conv1D_dense(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(conv1D_dense, self).__init__()
        if config is None:
            config = {
                'conv1d': {
                    'filter': 256,
                    'kernel': 2,
                    'activation': 'relu',
                    'padding': 'causal'
                },
                'dropout': 0.3,
                'pooling': 2,
                'dense': {
                    'activation': 'relu'
                }
            }
        self.mask = tkl.Masking(mask_value=value_miss)
        self.conv1 = tkl.Conv1D(
            filters=config['conv1d']['filter'],
            kernel_size=config['conv1d']['kernel'],
            activation=config['conv1d']['activation'],
            padding=config['conv1d']['padding'],

            dilation_rate=2
        )
        self.conv2 = tkl.Conv1D(
            filters=config['conv1d']['filter'],
            kernel_size=config['conv1d']['kernel'],
            activation=config['conv1d']['activation'],
            padding=config['conv1d']['padding']
        )
        self.conv3 = tkl.Conv1D(
            filters=config['conv1d']['filter'],
            kernel_size=config['conv1d']['kernel'],
            activation=config['conv1d']['activation'],
            padding=config['conv1d']['padding']
        )
        self.dropout = tkl.Dropout(config['dropout'])
        self.maxpool = tkl.MaxPooling1D(pool_size=config['pooling'])
        self.flatten = tkl.Flatten()
        self.dense = tkl.Dense(
            units=128,
            activation=config['dense']['activation']
        )
        self.dense1 = tkl.Dense(
            n_future * n_out_ft,
            activation=config['dense']['activation']
        )
        self.dense2 = tkl.Reshape((n_future, n_out_ft))

    def call(self, inputs, training=True, **kwargs):
        x = self.mask(inputs)
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.dropout(x)
        x = self.maxpool(x)
        x = self.flatten(x)
        x = self.dense(x)
        x = self.dense1(x)
        x = self.dense2(x)
        return x


class conv1D_lstm_dense(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(conv1D_lstm_dense, self).__init__()
        if config is None:
            config = {
                'conv1d': {
                    'filter': 256,
                    'kernel': 2,
                    'activation': 'relu',
                    'padding': 'causal'
                },
                'dropout': 0.3,
                'lstm': {
                    'units': 256,
                    'activation': 'relu'
                },
                'dense': {
                    'activation': 'relu'
                }
            }
        self.mask = tkl.Masking(mask_value=value_miss)
        self.conv1 = tkl.Conv1D(
            filters=config['conv1d']['filter'],
            kernel_size=config['conv1d']['kernel'],
            activation=config['conv1d']['activation'],
            padding=config['conv1d']['padding'],
            dilation_rate=2
        )
        self.conv2 = tkl.Conv1D(
            filters=config['conv1d']['filter'],
            kernel_size=config['conv1d']['kernel'],
            activation=config['conv1d']['activation'],
            padding=config['conv1d']['padding']
        )
        self.conv3 = tkl.Conv1D(
            filters=config['conv1d']['filter'],
            kernel_size=config['conv1d']['kernel'],
            activation=config['conv1d']['activation'],
            padding=config['conv1d']['padding']
        )
        self.dropout = tkl.Dropout(config['dropout'])
        self.lstm = tkl.LSTM(
            units=config['lstm']['units'],
            activation=config['lstm']['activation'],
        )
        self.dense = tkl.Dense(
            units=128,
            activation=config['dense']['activation']
        )
        self.dense1 = tkl.Dense(
            n_future * n_out_ft,
            activation=config['dense']['activation']
        )
        self.dense2 = tkl.Reshape((n_future, n_out_ft))

    def call(self, inputs, training=True, **kwargs):
        x = self.mask(inputs)
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.dropout(x)
        x = self.lstm(x)
        x = self.dense1(x)
        x = self.dense2(x)
        return x


class conv1D_bilstm_dense(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(conv1D_bilstm_dense, self).__init__()
        if config is None:
            config = {
                'conv1d': {
                    'filter': 64,
                    'kernel': 3,
                    'activation': 'relu',
                    'padding': 'causal'
                },
                'dropout': 0.3,
                'bilstm': {
                    'units': 256,
                    'activation': 'relu'
                },
                'dense': {
                    'activation': 'linear'
                }
            }
        self.mask = tkl.Masking(mask_value=value_miss)
        self.conv1 = tkl.Conv1D(
            filters=config['conv1d']['filter'],
            kernel_size=config['conv1d']['kernel'],
            activation=config['conv1d']['activation'],
            padding=config['conv1d']['padding'],
            dilation_rate=2
        )
        self.conv2 = tkl.Conv1D(
            filters=config['conv1d']['filter'],
            kernel_size=config['conv1d']['kernel'],
            activation=config['conv1d']['activation'],
            padding=config['conv1d']['padding']
        )
        self.conv3 = tkl.Conv1D(
            filters=config['conv1d']['filter'],
            kernel_size=config['conv1d']['kernel'],
            activation=config['conv1d']['activation'],
            padding=config['conv1d']['padding']
        )
        self.drop1 = tkl.Dropout(config['dropout'])
        self.bilstm = tkl.Bidirectional(
            tkl.LSTM(
                units=config['bilstm']['units'],
                activation=config['bilstm']['activation'],
            )
        )
        self.drop2 = tkl.Dropout(config['dropout'])
        self.flatten = tkl.Flatten()
        self.dense1 = tkl.Dense(
            n_future * n_out_ft,
            activation=config['dense']['activation']
        )
        self.dense2 = tkl.Reshape((n_future, n_out_ft))

    def call(self, inputs, training=True, **kwargs):
        x = self.mask(inputs)
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.drop1(x)
        x = self.bilstm(x)
        x = self.drop2(x)
        x = self.flatten(x)
        x = self.dense1(x)
        x = self.dense2(x)
        return x


class conv1D_bilstm_attention_dense(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(conv1D_bilstm_attention_dense, self).__init__()
        if config is None:
            config = {
                'conv1d': {
                    'filter': 64,
                    'kernel': 3,
                    'activation': 'relu',
                    'padding': 'causal'
                },
                'dropout': 0.3,
                'bilstm': {
                    'units': 256
                },
                'dense': {
                    'activation': 'linear'
                }
            }
        self.mask = tkl.Masking(mask_value=value_miss)
        self.conv1 = tkl.Conv1D(
            filters=config['conv1d']['filter'],
            kernel_size=config['conv1d']['kernel'],
            activation=config['conv1d']['activation'],
            padding=config['conv1d']['padding'],
            dilation_rate=2
        )
        self.drop1 = tkl.Dropout(config['dropout'])
        self.bilstm = tkl.Bidirectional(
            tkl.LSTM(
                units=config['bilstm']['units'],
                return_sequences=True
            )
        )
        self.drop2 = tkl.Dropout(config['dropout'])
        self.attention = attention()
        self.flatten = tkl.Flatten()
        self.dense1 = tkl.Dense(
            n_future * n_out_ft,
            activation=config['dense']['activation']
        )
        self.dense2 = tkl.Reshape((n_future, n_out_ft))

    def call(self, inputs, training=True, **kwargs):
        x = self.mask(inputs)
        x = self.conv1(x)
        x = self.drop1(x)
        x = self.bilstm(x)
        x = self.drop2(x)
        x = self.attention(x)
        x = self.flatten(x)
        x = self.dense1(x)
        x = self.dense2(x)
        return x



# -*- coding: utf-8 -*-
# @Time    : 29/08/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

def add_time_features(data, seasons: list = None):
    if seasons is None:
        seasons = ['Summer', 'Fall', 'Winter', 'Spring']
    data['minute'] = data.index.minute
    data['hour'] = data.index.hour
    data['day'] = data.index.day
    data['weekday'] = data.index.weekday
    data['month'] = data.index.month_name()
    data['year'] = data.index.year
    data['seasons'] = (
            (data.index.month % 12 + 3) // 3
    ).map(
        {
            1: seasons[0],
            2: seasons[1],
            3: seasons[2],
            4: seasons[3]
        }
    )
    data['day_name'] = data.index.day_name()
    data.loc[(data.index.weekday >= 5), 'day_type'] = 'non-working'
    data.loc[(data.index.weekday < 5), 'day_type'] = 'working'
    count = 0
    data = data.reset_index()
    data['date'] = data[data.columns[0]].apply(lambda x: x.strftime('%Y-%m-%d'))
    for i in data['date'].unique():
        data.loc[data['date'] == i, 'count_day'] = f'D{count}'
        count += 1
    #datasets = datasets.drop(['date'], axis=1)
    data = data.set_index(data.columns[0])
    return data


def print_summary(data):
    print('Start date:: ', data.index.min())
    print('Final date: ', data.index.max())
    print('Total number of days: ', data.index.max() - data.index.min())




# -*- coding: utf-8 -*-
# @Time    : 14/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

from openpy_fxts.base_tf import tkm, tkl
from openpy_fxts.architectures.layers.layers_class import type_rnn, dense_multi_out, rnn_drop
from openpy_fxts.architectures.layers.layers_class import multi_conv1d_pool_flat_drop, multi_rnn_drop
from openpy_fxts.architectures.layers.layers_class import dense_reshape_conv1d_multi_out
from openpy_fxts.architectures.layers.layers_class import multi_rnn_attention_conv1d, multi_birnn_multihead_attention_conv1d
from openpy_fxts.architectures.fxts.arch_dicts import _get_dict_fxts


class birnn_timedist_dense_class(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(birnn_timedist_dense_class, self).__init__()
        if config is None:
            config = _get_dict_fxts('BiRNN_TimeDist_Dense').config_birnn()
        self.n_future = n_future
        self.n_out_ft = n_out_ft
        self.config = config
        units = config['timedist_dense_layers']['units']
        activations = config['timedist_dense_layers']['activations']
        self.hidden_rnn = rnn_drop(
            mdl=type_rnn(config['rnn_layer']['type']),
            n_units=config['rnn_layer']['units'],
            activation=config['rnn_layer']['activations'],
            return_seq=config['rnn_layer']['sequences'],
            bidirectional=config['rnn_layer']['bidirectional'],
        )
        self.hidden_timedist_dense = []
        for j, (unit, act) in enumerate(zip(units, activations)):
            self.hidden_timedist_dense.append(tkl.TimeDistributed(tkl.Dense(units=unit, activation=act)))
        self.flatten = tkl.Flatten()
        self.hidden_dense = tkl.Dense(
            config['dense_layer']['units'],
            activation=config['dense_layer']['activation']
        )
        if config['dropout']['activate']:
            self.drop = tkl.Dropout(config['dropout']['rate'])
        self.layer_output = dense_multi_out(n_future, n_out_ft, config['output_layer']['activation'])

    def call(self, inputs, training=True, **kwargs):
        x = self.hidden_rnn(inputs)
        for layer in self.hidden_timedist_dense:
            x = layer(x)
        x = self.flatten(x)
        x = self.hidden_dense(x)
        if self.config['dropout']['activate']:
            x = self.drop(x)
        x = self.layer_output(x)
        return x


class birnn_conv1d_dense(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(birnn_conv1d_dense, self).__init__()
        if config is None:
            config = _get_dict_fxts('BiRNN_Conv1D').config_birnn()
        self.n_future = n_future,
        self.n_out_ft = n_out_ft,
        self.config = config
        self.multi_hidden_birnn = multi_rnn_drop(
            mdl=type_rnn(config['rnn_layers']['type']),
            units=config['rnn_layers']['units'],
            activations=config['rnn_layers']['activations'],
            sequences=config['rnn_layers']['sequences'],
            bidirectional=config['rnn_layers']['bidirectional'],
            use_drop=config['dropout']['activate'],
            rate=config['dropout']['rate']
        )
        self.multi_hidden_conv1d = multi_conv1d_pool_flat_drop(
            filters=config['conv1d_layers']['filters'],
            kernels=config['conv1d_layers']['kernels'],
            activations=config['conv1d_layers']['activations'],
            paddings=config['conv1d_layers']['paddings'],
            use_drop=config['dropout']['activate'],
            rate=config['dropout']['rate'],
            maxpooling_1d=True,
            pool_global=True,
            pool_size=config['conv1d_layers']['pool_size'],
            flatten=True)

        self.dense_conv1d_out = dense_reshape_conv1d_multi_out(
            n_future,
            n_out_ft,
            config['output_layer']['activation'],
            'same'
        )

    def call(self, inputs, training=True, **kwargs):
        x = self.multi_hidden_birnn(inputs)
        x = self.multi_hidden_conv1d(x)
        x = self.dense_conv1d_out(x)
        return x


class birnn_bahdanau_attention_conv1d_dense(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(birnn_bahdanau_attention_conv1d_dense, self).__init__()
        if config is None:
            config = _get_dict_fxts('BiRNN_Bahdanau_Attention_Conv1D').config_birnn()
        self.config = config
        self.multi_hidden_layers = multi_rnn_attention_conv1d(
            mdl=type_rnn(config['rnn_layers']['type']),
            bidirectional=config['rnn_layers']['bidirectional'],
            units=config['rnn_layers']['units'],
            rnn_act=config['rnn_layers']['activations'],
            return_seq=config['rnn_layers']['sequences'],
            att_type='Bahdanau',
            kernel_reg=config['attention']['kernel_reg'],
            bias_reg=config['attention']['bias_reg'],
            reg_weight=config['attention']['reg_weight'],
            n_filters=config['conv1d_layers']['filters'],
            kernels=config['conv1d_layers']['kernels'],
            conv_act=config['conv1d_layers']['activations'],
            paddings=config['conv1d_layers']['paddings'],
            simple_pool=True,
            global_pool=False,
            pool_size=config['conv1d_layers']['pool_size']
        )
        self.flatten = tkl.Flatten()
        if config['dropout']['activate']:
            self.dropout = tkl.Dropout(config['dropout']['rate'])
        self.dense_out = dense_multi_out(
            n_future,
            n_out_ft,
            config['output_layer']['activation']
        )

    def call(self, inputs, training=True, **kwargs):
        x = self.multi_hidden_layers(inputs)
        x = self.flatten(x)
        if self.config['dropout']['activate']:
            x = self.dropout(x)
        x = self.dense_out(x)
        return x


class birnn_luong_attention_conv1d_dense(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(birnn_luong_attention_conv1d_dense, self).__init__()
        if config is None:
            config = _get_dict_fxts('BiRNN_Luong_Attention_Conv1D').config_birnn()
        self.config = config
        self.multi_hidden_layers = multi_rnn_attention_conv1d(
            mdl=type_rnn(config['rnn_layers']['type']),
            bidirectional=config['rnn_layers']['bidirectional'],
            units=config['rnn_layers']['units'],
            rnn_act=config['rnn_layers']['activations'],
            return_seq=config['rnn_layers']['sequences'],
            att_type='Luong',
            kernel_reg=config['attention']['kernel_reg'],
            bias_reg=config['attention']['bias_reg'],
            reg_weight=config['attention']['reg_weight'],
            n_filters=config['conv1d_layers']['filters'],
            kernels=config['conv1d_layers']['kernels'],
            conv_act=config['conv1d_layers']['activations'],
            paddings=config['conv1d_layers']['paddings'],
            simple_pool=True,
            global_pool=False,
            pool_size=config['conv1d_layers']['pool_size']
        )
        self.flatten = tkl.Flatten()
        if config['dropout']['activate']:
            self.dropout = tkl.Dropout(config['dropout']['rate'])
        self.dense_out = dense_multi_out(
            n_future,
            n_out_ft,
            config['output_layer']['activation']
        )

    def call(self, inputs, training=True, **kwargs):
        x = self.multi_hidden_layers(inputs)
        x = self.flatten(x)
        if self.config['dropout']['activate']:
            x = self.dropout(x)
        x = self.dense_out(x)
        return x


class birnn_multihead_attention_conv1d_dense(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(birnn_multihead_attention_conv1d_dense, self).__init__()
        if config is None:
            config = _get_dict_fxts('BiRNN_MultiHeadAttention_Conv1D').config_birnn()
        self.config = config
        self.multi_hidden_layers = multi_birnn_multihead_attention_conv1d(
            mdl=type_rnn(config['rnn_layers']['type']),
            bidirectional=config['rnn_layers']['bidirectional'],
            units=config['rnn_layers']['units'],
            rnn_act=config['rnn_layers']['activations'],
            return_seq=config['rnn_layers']['sequences'],
            att_type='MultiHeadAttention',
            num_heads=config['attention']['num_heads'],
            key_dim=config['attention']['key_dim'],
            dropout=config['attention']['dropout'],
            n_filters=config['conv1d_layers']['filters'],
            kernels=config['conv1d_layers']['kernels'],
            conv_act=config['conv1d_layers']['activations'],
            paddings=config['conv1d_layers']['paddings'],
            simple_pool=True,
            global_pool=False,
            pool_size=config['conv1d_layers']['pool_size']
        )
        self.flatten = tkl.Flatten()
        if config['dropout']['activate']:
            self.dropout = tkl.Dropout(config['dropout']['rate'])
        self.dense_out = dense_multi_out(
            n_future,
            n_out_ft,
            config['output_layer']['activation']
        )

    def call(self, inputs, training=True, **kwargs):
        x = self.multi_hidden_layers(inputs)
        x = self.flatten(x)
        if self.config['dropout']['activate']:
            x = self.dropout(x)
        x = self.dense_out(x)
        return x






# -*- coding: utf-8 -*-
# @Time    : 24/03/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

import logging
import pandas as pd
import numpy as np
from sklearn.metrics import mean_absolute_error, mean_squared_error, mean_absolute_percentage_error, r2_score
from openpy_fxts.models.imputation.base_lib import init_class
from openpy_fxts.error_handling_logging import update_logg_file

log_py = logging.getLogger(__name__)


class Metrics:

    def add_dict_metrics(
            self,
            df_true,
            df_imp,
            dict_perf,
            method
    ):
        df_aux = pd.DataFrame()
        for x in range(df_true.shape[1]):
            y_true = np.reshape(np.array(df_true.iloc[:, x]), ((df_true.iloc[:, x].shape[0], 1)))
            y_imp = np.reshape(np.array(df_imp.iloc[:, x]), ((df_imp.iloc[:, x].shape[0], 1)))
            dict_metrics = self._metrics(y_true, y_imp)
            row_to_append = pd.DataFrame([dict(zip(list(dict_metrics.keys()), list(dict_metrics.values())))])
            df_aux = pd.concat([df_aux, row_to_append])
        df_aux.insert(0, 'Index', list(df_true.columns))
        df_aux.set_index('Index', inplace=True)
        dict_perf[method] = df_aux
        return dict_perf

    def _metrics(self, y_true, y_imp):
        metrics = dict()
        metrics['MAE'] = mean_absolute_error(y_true, y_imp)
        metrics['MSE'] = mean_squared_error(y_true, y_imp)
        metrics['MAPE'] = mean_absolute_percentage_error(y_true, y_imp)
        metrics['R2'] = r2_score(y_true, y_imp)
        return metrics


class Utils(init_class):
    def __init__(self, dataframe=None):
        super().__init__(dataframe)
        self.dataframe = dataframe

    def list_column_missing(self):
        data_NaN = pd.DataFrame(self.dataframe.isnull().sum(), columns=['number'])
        return list(data_NaN[data_NaN['number'] > 0].index)

    def check_missing(self):
        # summarize the number of rows with missing values for each column
        aux = self.dataframe.isnull().values.any()
        return aux

    def missing_values_table(self):
        # Total missing values
        mis_val = self.dataframe.isnull().sum()

        # Percentage of missing values
        mis_val_percent = 100 * self.dataframe.isnull().sum() / len(self.dataframe)

        # Make a table with the results
        mis_val_table = pd.concat([mis_val, mis_val_percent], axis=1)

        # Rename the columns
        mis_val_table_ren_columns = mis_val_table.rename(
            columns={0: 'Missing Values', 1: '% of Total Values'})

        # Sort the table by percentage of missing descending
        mis_val_table_ren_columns = mis_val_table_ren_columns[mis_val_table_ren_columns.iloc[:, 1] != 0].sort_values(
            '% of Total Values', ascending=False).round(1)

        # Print some summary information
        update_logg_file(
            "Your selected DataFrame has " + str(self.dataframe.shape[1]) + " columns.\n"
            "There are " + str(mis_val_table_ren_columns.shape[0]) +
            " columns that have missing values.", 0, log_py
        )

        # Return the dfframe with missing information
        return mis_val_table_ren_columns

    def total_days_range(self):
        print('Start date: ', self.dataframe.index.min())
        print('Final date: ', self.dataframe.index.max())
        print('Total number of Days: ', self.dataframe.index.max() - self.dataframe.index.min())

    def duplicate(self):
        pass


class ft_eng(init_class):

    def __init(self, dataframe):
        super().__init__(dataframe)
        self.dataframe = dataframe

    def add_datetime(self):
        self.dataframe['Year'] = self.dataframe.index.year
        self.dataframe['Month'] = self.dataframe.index.month
        self.dataframe['Day'] = self.dataframe.index.day
        self.dataframe['nDay'] = self.dataframe.index.weekday
        self.dataframe['Hour'] = self.dataframe.index.hour
        self.dataframe['Minute'] = self.dataframe.index.minute

    def add_seasons(self):
        self.dataframe.loc[
            ((self.dataframe["Month"] == 12) & (self.dataframe["Day"] >= 21)) | (self.dataframe["Month"] == 1) | (
                        self.dataframe["Month"] == 2) | (
                    (self.dataframe["Month"] == 3) & (self.dataframe["Day"] <= 20)), 'Season'] = 'Summer'
        self.dataframe.loc[
            ((self.dataframe["Month"] == 3) & (self.dataframe["Day"] >= 21)) | (self.dataframe["Month"] == 4) | (
                        self.dataframe["Month"] == 5) | (
                    (self.dataframe["Month"] == 6) & (self.dataframe["Day"] <= 20)), 'Season'] = 'Fall'
        self.dataframe.loc[
            ((self.dataframe["Month"] == 6) & (self.dataframe["Day"] >= 21)) | (self.dataframe["Month"] == 7) | (
                        self.dataframe["Month"] == 8) | (
                    (self.dataframe["Month"] == 9) & (self.dataframe["Day"] <= 20)), 'Season'] = 'Winter'
        self.dataframe.loc[
            ((self.dataframe["Month"] == 9) & (self.dataframe["Day"] >= 21)) | (self.dataframe["Month"] == 10) | (
                        self.dataframe["Month"] == 11) | (
                    (self.dataframe["Month"] == 12) & (self.dataframe["Day"] <= 20)), 'Season'] = 'Spring'

    def add_working_day(self):
        self.dataframe.loc[self.dataframe["nDay"] >= 5, 'Day type'] = 'non-working'
        self.dataframe.loc[(self.dataframe["nDay"] < 5), 'Day type'] = 'working'


class Preprocessing:

    def LabelEncoder(self, df):
        transform_dict = {}
        for col in df.iloc[:, :].select_dtypes(include=['object']).columns:
            cats = pd.Categorical(df[col]).categories
            d = {}
            for i, cat in enumerate(cats):
                d[cat] = i
            transform_dict[col] = d

        return df.replace(transform_dict), transform_dict

    def inverse_LabelEncoder(self, df, transform_dict):
        inverse_transform_dict = {}
        for col, d in transform_dict.items():
            inverse_transform_dict[col] = {v: k for k, v in d.items()}

        return df.replace(inverse_transform_dict)

    def index_column(self, df):
        dict = {}
        aux = 0
        for i in df.columns:
            dict[i] = aux
            aux += 1
        list_aux = list()
        for col in df.iloc[:, :].select_dtypes(include=['object']).columns:
            list_aux.append(col)
        list_index = list()
        for m in list_aux:
            list_index.append(dict[m])
        return list_index
